#include <math.h>

// Very small value
#define EPSI 0.00000001

// Frequency 365/(2xPI)
#define FREQ 58.091554

// PI
#define PI 3.141592653

/***********************************************
* Conversion of annual Temperature values in ETP time series (non dated values)
*
*	c_ETP inputs :
*	nval	= number of required values
*	PAS	= Time step length (in hours, PAS <= 24 - max daily values)
*	LAT	= Latitude of the station location
*	TEMPM	= 12 mean mensual temperature values (matrix 12*1)
*	DATEDEP	= Starting date of the ETP data (format AAAAMMJJHHmm)
*	TYPE    = Type of input values (T = 1, directly monthly ETP values = 2)
*
*	output :
*	ETP 	= PE time series
*
************************************************/

/* --- Time management functions and subs (cf. c_CONVDATE) -- */
void c_CONVDATE(long int nval,int * FORMAT,double * DATES,double * CONV);


/* --- Fourrier decomposition subroutines for 12 monthly values (4 harmonics) ---- 
	Signal = Input signal to decompose in Fourrier harmonics
	fCoeff = Fourrier coefficients
*/
void FOUR(double d_in_m[],double Signal[],double fCoeff[])
{
	int i,j,k,l;
	double Diag,Diag2;
	double fArg[9][12], fArg2[9][9], sumD[12];
	double fCoeffInteg[9];

	// Julian days in a year (non bisextile)
	sumD[0]=d_in_m[0];
	for(i=1;i<12;i++){sumD[i]=sumD[i-1]+d_in_m[i];}


	// Arguments of the fourrier decomposition
	fArg[0][0]=d_in_m[0];
	fArg[1][0]=sin(sumD[0]/FREQ);
	fArg[2][0]=cos(sumD[0]/FREQ)-1;
	fArg[3][0]=sin(sumD[0]/FREQ*2);
	fArg[4][0]=cos(sumD[0]/FREQ*2)-1;
	fArg[5][0]=sin(sumD[0]/FREQ*3);
	fArg[6][0]=cos(sumD[0]/FREQ*3)-1;
	fArg[7][0]=sin(sumD[0]/FREQ*4);
	fArg[8][0]=cos(sumD[0]/FREQ*4)-1;

	for(j=1;j<12;j++){
		fArg[0][j]=d_in_m[j];
		fArg[1][j]=sin(sumD[j]/FREQ)-sin(sumD[j-1]/FREQ);
		fArg[2][j]=cos(sumD[j]/FREQ)-cos(sumD[j-1]/FREQ);
		fArg[3][j]=sin(sumD[j]/FREQ*2)-sin(sumD[j-1]/FREQ*2);
		fArg[4][j]=cos(sumD[j]/FREQ*2)-cos(sumD[j-1]/FREQ*2);
		fArg[5][j]=sin(sumD[j]/FREQ*3)-sin(sumD[j-1]/FREQ*3);
		fArg[6][j]=cos(sumD[j]/FREQ*3)-cos(sumD[j-1]/FREQ*3);
		fArg[7][j]=sin(sumD[j]/FREQ*4)-sin(sumD[j-1]/FREQ*4);
		fArg[8][j]=cos(sumD[j]/FREQ*4)-cos(sumD[j-1]/FREQ*4);
	}

	// Calculation of fArg x fArg (Rowise product)
	for(i=0;i<9;i++){
		for(j=0;j<9;j++){
			fArg2[i][j]=0;
			for(k=0;k<12;k++){
				fArg2[i][j]+=fArg[i][k]*fArg[j][k];
			}			
		}
	}

	// fArg2 normalisation
	for(k=0;k<9;k++){
		// Diagonal term of the fArg2 matrix
		Diag=fArg2[k][k];
		fArg2[k][k]=1;

		if(Diag==0){return;} // Fourrier decomposition impossible !!
		
		// Normalisation of each line of the matrix by the diagonal term
		for (l=0;l<9;l++){
			fArg2[k][l]=fArg2[k][l]/Diag;
		}

		// Second step of the normalisation
		for (i=0;i<9;i++){
			if(i!=k){
				Diag2=fArg2[i][k];
				fArg2[i][k]=0;
				for(j=0;j<9;j++){fArg2[i][j]-=Diag2*fArg2[k][j];}
			}
		}
	}

	// Calculation of the Fourrier coefficients by convolution
	for(i=0;i<9;i++){
		fCoeffInteg[i]=0;
		for(j=0;j<12;j++){
			fCoeffInteg[i]+=fArg[i][j]*Signal[j]; // Signal must have 12 elements !
		}
	}

	// Final integration
	for(i=0;i<9;i++){
		// Initialisation
		fCoeff[i]=0;
		for(j=0;j<9;j++){
			fCoeff[i]+=fArg2[i][j]*fCoeffInteg[j];
		}
	}	
}

/* --- Rayonnement extraterrestre------------ 
*	JJ : jour julien
*	LAT: Latitude (radians)
*/
double RAYONEXT(double JJ, double LAT)
{
	// ATTENTION PB si LAT=0 ou teta = 0 !!!
		
	double teta, eta, cosfi, costeta, cosgz, cosom, cosom2, sinom,om, cospz;
	
	cosfi = cos(LAT);
	teta  = 0.4093*sin(JJ/58.1-1.405);

	costeta = cos(teta);
	if(cos(LAT-teta)>0.001){cosgz = cos(LAT-teta);}
	else {cosgz=0.001;}
	if((cosfi!=0) & (costeta!=0)){cosom = 1-cosgz/cosfi/costeta;}
	else{cosom=1;}

	if(cosom>1){cosom=1;}
	if(cosom<-1){cosom=-1;}
	cosom2=pow(cosom,2);

	if(cosom2>=1){sinom=0;}
	else {sinom = sqrt(1-cosom2);}
	om = acos(cosom);

	if(om!=0){cospz=cosgz+cosfi*costeta*(sinom/om-1);}
	else{;}
	if(cospz<0.001){cospz = 0.001;}
	eta=1+cos(JJ/58.1)/30;

	return 15.2*om*cospz*eta;
}


/* ---------- Main sub ----------------------- */

void c_ETP(long int nval,double PAS[], double LAT[], double TEMPM[], double DATEDEP[], double TYPEINPUT[], double ETP[])
{
	int i,j,m,ts,FORMAT[1],annCour,moCour,joCour,pascompte,TypeInput;
	double JJ1,JJ2,lat,pas,dateDep,dateCour;
	double DATES[1],CONV[1];
	double TJ[31][12],ETPJJ[31][12],d_in_m[12],TEMPinteg[12],fCoeff[9];

	double DISTRIB[1440],distrib1,distrib2,ETPJ; // PAS > 1 min !

	// Initialisations
	JJ1=0;
	JJ2=0;

	TypeInput=(int)(TYPEINPUT[0]);
	if(!(TypeInput==1)&!(TypeInput==2)){TypeInput=1;}

	// Time step must provide a 
	pas=PAS[0];
	if((pas<24)&(floor(24/pas)!=24/pas)){for(ts=0;ts<nval;ts++){ETP[ts]=-9.999;return;}} 

	if(LAT[0]>0){lat=LAT[0];}

	FORMAT[0]=0;
	if(DATEDEP[0]>0){
		c_CONVDATE(1,FORMAT,DATEDEP,CONV);
		if(CONV[0]==-999){dateDep=1;}
		else{dateDep=CONV[0];}
	}
	FORMAT[0]=1;

	for(i=0;i<31;i++){for(j=0;j<12;j++){ETPJJ[i][j]=0;TJ[i][j]=0;}}

	d_in_m[0]=31;
	d_in_m[1]=28;	
	d_in_m[2]=31;
	d_in_m[3]=30;
	d_in_m[4]=31;
	d_in_m[5]=30;
	d_in_m[6]=31;
	d_in_m[7]=31;
	d_in_m[8]=30;
	d_in_m[9]=31;
	d_in_m[10]=30;
	d_in_m[11]=31;

	for(i=0;i<12;i++){TEMPinteg[i]=TEMPM[i]*d_in_m[i];}


	/*--- Parameter control ----------
	printf("\nnval=%5d npas=%5.2f lat=%5.2f dateDep=%5.2f\n",nval,pas,lat,dateDep);
	for(i=0;i<12;i++){printf("TEMPinteg[%2d]=%8.2f\n",i,TEMPinteg[i]);}
	*/	

	// Distribution of ETP during the day for time-steps inferior to 24 hours
	distrib1=0;
	distrib2=pas;

	if(pas<24){
		for(i=0;i<(int)(24/pas);i++){
			DISTRIB[i]=(12/PI*(cos(PI*distrib1/12-PI/2)-cos(PI*distrib2/12-PI/2))+pas)/24;
			distrib1+=pas;
			distrib2+=pas;
		}
	}

	// Fourrier coefficients calculation
	FOUR(d_in_m,TEMPinteg,fCoeff);

	/* -- Control of fCoeff content --
	for(j=0;j<9;j++){printf("\nfCoeff[%d]=%10.4f",j,fCoeff[j]);}
	printf("\n");
	*/

	// Generation of the 31 x 12 matrix containing the daily ETP values (matrix ETPJ)
	for(m=0;m<12;m++){

		for(i=0;i<(int)(d_in_m[m]);i++){
			JJ2=JJ1+1;

			// Application of the Fourrier developement
			TJ[i][m]=fCoeff[0];
			TJ[i][m]+=fCoeff[1]*(sin(JJ2/FREQ)-sin(JJ1/FREQ));
			TJ[i][m]+=fCoeff[2]*(cos(JJ2/FREQ)-cos(JJ1/FREQ));
			TJ[i][m]+=fCoeff[3]*(sin(JJ2/FREQ*2)-sin(JJ1/FREQ*2));
			TJ[i][m]+=fCoeff[4]*(cos(JJ2/FREQ*2)-cos(JJ1/FREQ*2));	
			TJ[i][m]+=fCoeff[5]*(sin(JJ2/FREQ*3)-sin(JJ1/FREQ*3));
			TJ[i][m]+=fCoeff[6]*(cos(JJ2/FREQ*3)-cos(JJ1/FREQ*3));
			TJ[i][m]+=fCoeff[7]*(sin(JJ2/FREQ*4)-sin(JJ1/FREQ*4));
			TJ[i][m]+=fCoeff[8]*(cos(JJ2/FREQ*4)-cos(JJ1/FREQ*4));
			
			// Application of the ETP calculation
			if(TJ[i][m]<=-5){ETPJJ[i][m]=0;}
			else{
				if(TypeInput==1){ETPJJ[i][m]=RAYONEXT(JJ2,lat)*(TJ[i][m]+5)/100;}
				if(TypeInput==2){ETPJJ[i][m]=TJ[i][m]/d_in_m[m];} // ETP is provided directly
			}

			// Julian day increment
			JJ1=JJ2;
		}
	}

	// 29th february ETP calculation
	ETPJJ[28][1]= 0.5*(ETPJJ[27][1]+ETPJJ[0][2]);

	// Generation of the ETP time series
	for(ts=0;ts<nval;ts++){

		// Get the current date in calendar format
		DATES[0] = dateDep+ts*pas/24;

		if(DATES[0]==floor(DATES[0])){
			c_CONVDATE(1,FORMAT,DATES,CONV);
			dateCour = CONV[0];
			annCour=(int)(dateCour/100000000);
			moCour=(int)((dateCour)/1000000-annCour*100);
			joCour=(int)((dateCour)/10000-annCour*10000-moCour*100);

			// daily ETP calculation (daily values only)
			ETPJ=ETPJJ[joCour-1][moCour-1];

			pascompte=0;
		}
		else{pascompte++;}
				
		
		// Storage of ETP value
		if(pas<24){ETP[ts]=ETPJ*DISTRIB[pascompte];}
		else{ETP[ts]=ETPJ;}

		/* -- Control of ETP content -- 
		printf("j=%2d m=%2d (%5d) - %16.5f - %8.5f %8.5f\n",joCour,moCour,pascompte,DATES[0],ETPJ,ETP[ts]);
		*/

	}
	
}
