#include <math.h>
#include <sciprint.h>

// Number of season per year
#define NBSEAS 6

// Size of the PX vector (= (NBSEAS+1) * 8)
#define SIZEPX 56

// Size of rainfall vector (200 years of rainfall)
#define SIZE 75000

#define EPS 0.00000003
#define PI 3.14159265

/***********************************************
* Calculation of parameters for the daily rainfall generator (c_GENEPLUIE)
*	3 states (x0 =mean(P,P>0) x log(2) ) : 
*		(1) P=0
*		(2) P>0 & P<=x0
*		(3) P>x0
*
* Repartition function (u = random number in ]0..1[ )
*		(1) P = 0 (no repartition function needed)
*		(2) P = [exp(b.x/x0)-1]/[exp(b)-1]
*		(3) P = 1 - exp[(x-x0)/a]
*
* a and b are generator parameters for each of the 6 seasons (every two months)
*		a = mean(P,P>x0) // to preserve the mean values of strong rainfall values
*		b =      	 // to preserve the mean values of low rainfall values
*
*	DATEDEP[0]	Starting date (format 199901010000)
*       Pinp		Measured rainfall time series (daily values)
*	PX		Parameters of the rainfall generator (matrix [NBSEAS+1]x8) 
*			PX(i,j) = generator parameter (i =[1..NBSEAS], j=[1..8]) calculated from rainfall time series.
*
*				PX(1,:) = a parameter >> average of the low rainfall values (0<..<=Pmoy)
*				PX(8,:) = b parameter >> see above

*				PX(2,:) = probability to have a state 1 with a state 1 on preceeding day	
*				PX(4,:) = probability to have a state 1 with a state 2 on preceeding day	
*				PX(6,:) = probability to have a state 1 with a state 3 on preceeding day	
*				PX(3,:) = probability to have a state 2 with a state 1 on preceeding day	
*				PX(5,:) = probability to have a state 2 with a state 2 on preceeding day	
*				PX(7,:) = probability to have a state 2 with a state 3 on preceeding day	
*
*				PX(NBSEAS+1,1)=average of positive rainfall x log(2)
*				PX(NBSEAS+1,2)=average of rainfall
*
************************************************/

/* --- Time management functions and subs (cf. c_GENEPLUIE) -- */
extern void DEMAIN(int * j,int * m,int * a);

/** Truncated exponential ****** */
double TRONC(double PMS)
{
	double A,C,RES,sign;

	sign=1;
      	A=12*(PMS-0.5);

	if(fabs(A)<EPS){return A;}
	else{

		// Code de Claude
		//if(PMS<0.5){A=(3*PMS-1-pow(1+3*PMS*(2-5*PMS),0.5))/2/PMS;}
		//else{A=(3*PMS-2+pow(3*PMS*(8-PMS)-8,0.5))/2/(1-PMS);}

		if(PMS>=0.5){sign=-1;PMS=1-PMS;}

		A=(3*PMS-1-pow(1+3*PMS*(2-5*PMS),0.5))/2/PMS;
		C=exp(-1*A);
		RES=A-(1/(1-C)-1/A-PMS)/(1/A/A-C/pow(1-C,2));

		return sign*RES;
	}	
	
}


/***** Routine to correct the parameters of the generator ************/

void COMP(double * param)
{
	int i,k,KM,KP,LY[NBSEAS];
	// Loop over the 8 parameters
	for(i=0;i<8;i++)
	{
		// First loop over the seasons to look for missing values in the parameters table
		for(k=0;k<NBSEAS;k++)
		{
			LY[k]=0;
			if(param[k+(NBSEAS+1)*i]==-9.999){LY[k]=1;}
		}

		// Second loop over the seasons to check if there are sufficient parameters to interpolate
		for(k=0;k<NBSEAS;k++)
		{
			KM=k+1;if(KM==NBSEAS){KM=0;}				
			if((LY[k]==1)&(LY[KM]==1)){return;} // Not enough parameters
		}

		// Third loop over the seasons to interpolate if possible
		for(k=0;k<NBSEAS;k++)
		{
			if(LY[k]==1)
			{
				KP=k+1;if(KP==NBSEAS){KP=0;}				
				KM=k-1;if(KM==-1){KM=NBSEAS-1;}								
				param[k+(NBSEAS+1)*i]=(param[KP+(NBSEAS+1)*i]+param[KM+(NBSEAS+1)*i])/2;
			}
		}
	}
}

/***** CORE CODE ************/
void c_GENEPAR( int nval, double * DATEDEP,double * Pinp, 
                double * MULTPPOS, double * PX)
{
	int i,j,k,SEAS,valpos,valstrictpos,LAC,NBLOWLY;
	double PU,PP,LY[SIZEPX],PK,mult,MAXP;
	int ann,moi,jou;

  // Verification of MULTPPOS
  mult = MULTPPOS[0];
  if(mult>1){mult=log(2);sciprint("\tMULTPPOS is > 1 ! Converted to log(2)\n");}

	// Conversion of the simulation starting date 
	// Identification of the year, month and day in the month of the beginning of the simulation
	if(DATEDEP[0]>0){
		ann=(int)(DATEDEP[0]/100000000);moi=(int)(DATEDEP[0]/1000000-ann*100);
		jou= (int)(DATEDEP[0]/10000-ann*10000-moi*100);
	}
	// Initialisation of PX and LY
	for(i=0;i<NBSEAS+1;i++){
		for(j=0;j<8;j++){PX[i+(NBSEAS+1)*j]=0;LY[i+(NBSEAS+1)*j]=0;}
	}

	// Calculation of the mean positive rainfall x MULTP0 !!!!!
	valpos=0;valstrictpos=0;
	for(i=0;i<nval-1;i++){
	 PP=Pinp[i];
		if(PP>=0){PX[2*NBSEAS+1]+=PP;valpos++;}
		if(PP>0){PX[NBSEAS]+=PP;valstrictpos++;}
	}
	if(valpos>0){PX[2*NBSEAS+1]/=valpos;}
	else {PX[2*NBSEAS+1]=-9.999;}
	if(valstrictpos>0){PX[NBSEAS]*=mult/valstrictpos;}
	else {PX[NBSEAS]=-9.999;}

	//printf("Pluie Moy: PX[%d]=%7.4f\n\n",NBSEAS,PX[NBSEAS]);

	// First calculation  of parameters  (over the rainfall time series)
	MAXP=0;
	for(i=0;i<nval-1;i++){    
		// Identification of the current year, month and day in the month
		DEMAIN(&jou,&moi,&ann);

		// Season number
		SEAS = 1 + (int)((moi-1)*NBSEAS/12);
		//printf("%5d - DATES[0]=%5.0f SEAS=%d\n",i,DATES[0],SEAS);

		// Rainfall value of the current day and the following day
		PP = Pinp[i];PU = Pinp[i+1];

    // Maximum value
    if(PP>MAXP){MAXP=PP;}

		// Calculation of the propababilities with all the possible cases :
		// PP = 0 / 0-Pmoy / > Pmoy  & PU = 0 / 0-Pmoy / > Pmoy  >> 9 cases but only 6 treated
		if((PU>=0)&(PP>=0))
		{
			if(PP==0){
				LY[SEAS-1]+=1;
				if(PU==0){PX[SEAS-1+(NBSEAS+1)]+=1;}              // On incrmente PAR(sais,2)
				if(PU<=PX[NBSEAS]){PX[SEAS-1+(NBSEAS+1)*2]+=1;}   // On incrmente PAR(sais,3)
			}
			if((PP>0)&(PP<=PX[NBSEAS])){
				LY[SEAS-1+(NBSEAS+1)]+=1;
				PX[SEAS-1+(NBSEAS+1)*7]+=PP;
				if(PU==0){PX[SEAS-1+(NBSEAS+1)*3]+=1;}            // On incrmente PAR(1,4)
				if(PU<=PX[NBSEAS]){PX[SEAS-1+(NBSEAS+1)*4]+=1;}	  // On incrmente PAR(1,5)			
			}
			if(PP>PX[NBSEAS]){
				LY[SEAS-1+(NBSEAS+1)*2]+=1;
				PX[SEAS-1]+=PP-PX[NBSEAS];
				if(PU==0){PX[SEAS-1+(NBSEAS+1)*5]+=1;}
				if(PU<=PX[NBSEAS]){PX[SEAS-1+(NBSEAS+1)*6]+=1;}				
			}
		}
	}
  
  // Sciprint of contengency table

  sciprint("\tMean Rainfall      = %6.2f\n",PX[2*NBSEAS+1]);
  sciprint("\tP0 (threshold 1/2) = %6.2f\n",PX[NBSEAS]);
  sciprint("\tMaximum Rainfall   = %6.2f\n",MAXP);
  sciprint("\n\t***************\n");
  sciprint("\tContengency table (number of values): \n\n");
  sciprint("\t%6s %6s %6s %6s %6s %6s %6s %6s %6s %6s\n","",
            "N1","(1|1)","(1|2)",
            "N2","(2|1)","(2|2)",
            "N3","(3|1)","(3|2)");
            
  NBLOWLY=0;
  for(SEAS=1;SEAS<NBSEAS+1;SEAS++){
    sciprint("\tSeas %d",SEAS);
    sciprint(" %6.0f %6.0f %6.0f %6.0f %6.0f %6.0f %6.0f %6.0f %6.0f",
      LY[SEAS-1],PX[SEAS-1+NBSEAS+1],PX[SEAS-1+(NBSEAS+1)*2]-PX[SEAS-1+NBSEAS+1],
      LY[SEAS-1+NBSEAS+1],PX[SEAS-1+(NBSEAS+1)*3],PX[SEAS-1+(NBSEAS+1)*4]-PX[SEAS-1+(NBSEAS+1)*3],
      LY[SEAS-1+(NBSEAS+1)*2],PX[SEAS-1+(NBSEAS+1)*5],PX[SEAS-1+(NBSEAS+1)*6]-PX[SEAS-1+(NBSEAS+1)*5]);
    sciprint("\n");
    if(PX[SEAS-1+(NBSEAS+1)*5]<4){NBLOWLY++;}
  } 
  sciprint("\n");

  // Warning if mult is too large
  if(NBLOWLY>2){
    sciprint("\n\t***************\n\n");
    sciprint("\tWARNING ! %d seasons have a very low P(3|1), MULTPPOS should be decreased\n\n",
        NBLOWLY);
  }

  
	// Final values
	LAC=0;

	for(k=0;k<NBSEAS;k++){
		// First state parameters
		if(LY[k]>5){
			PX[k+(NBSEAS+1)]=PX[k+(NBSEAS+1)]/LY[k];
			PX[k+(NBSEAS+1)*2]=PX[k+(NBSEAS+1)*2]/LY[k];
		}
		else{LAC=1;PX[k+(NBSEAS+1)]=-9.999;PX[k+(NBSEAS+1)*2]=-9.999;}

		// Second state parameters
		if(LY[k+(NBSEAS+1)]>5){
			PX[k+(NBSEAS+1)*3]=PX[k+(NBSEAS+1)*3]/LY[k+(NBSEAS+1)];
			PX[k+(NBSEAS+1)*4]=PX[k+(NBSEAS+1)*4]/LY[k+(NBSEAS+1)];
			PK=PX[k+(NBSEAS+1)*7]/LY[k+(NBSEAS+1)]/PX[NBSEAS];
			PX[k+(NBSEAS+1)*7]=TRONC(PK);
		}
		else{
			LAC=1;
			PX[k+(NBSEAS+1)*3]=-9.999;
			PX[k+(NBSEAS+1)*4]=-9.999;
			PX[k+(NBSEAS+1)*7]=-9.999;
		}

		// Third state parameters
		if(LY[k+(NBSEAS+1)*2]>5){
			PX[k]=PX[k]/LY[k+(NBSEAS+1)*2];
			PX[k+(NBSEAS+1)*5]=PX[k+(NBSEAS+1)*5]/LY[k+(NBSEAS+1)*2];
			PX[k+(NBSEAS+1)*6]=PX[k+(NBSEAS+1)*6]/LY[k+(NBSEAS+1)*2];
		}
		else{
			LAC=1;PX[k]=-9.999;
			PX[k+(NBSEAS+1)*5]=-9.999;
			PX[k+(NBSEAS+1)*6]=-9.999;
		}
	}

	// Interpolation of parameters if there are missing values
	if(LAC==1){COMP(PX);}
	
}
