#include <math.h>

// Number of tolerated missing data
#define NLACMAX 5

// Number of parameters
#define	npX 5


/***********************************************
* GR2M model
*	PAS = Time step (hours)	
*
*	GR2M Parameters :
*	X[0] = Soil reservoir capacity 		(> 1mm)
*	X[1] = exchange parameter 		(in [0..50])
*	X[2] = routing store capacity (fixed to 60 mm)
*	X[3] = delay (not used)
*	X[4] = second exchange parameter (not used)
* Ref :
* Mouelhi, S., Michel, C., Perrin, C., Andreassian, V.
* Stepwise development of a two-parameter monthly water balance model 
* (2006) Journal of Hydrology, 318 (1-4), pp. 200-214. 
************************************************/


/* ************** Core subroutine ****************************** */

void c_GR2M(int nval,double * PAS,double * Xinp,double * Pinp,
	double * Einp,double * Rcalc,double * Scalc,double * Qcalc, 
	double * INI, double * ECHcalc, double * PRcalc)
{
	int i,nLAC;
	double P,E,Q,Sini,Rini,TPr,IGF,phi,psi; 
	double X[5],V[2];
	double S1,S2,P1,P2,P3,ECH,R2,R1,Q2,PERCFACTOR,pas;


	/* --- Model structure ----------------- */

	//Time step
	if (PAS[0]<1) {pas=1;} 
	else {pas=PAS[0];}

	// Percolation factor
	if (pas<12) 	{PERCFACTOR=4;} 	// Hourly time steps
	if (pas>=12) {PERCFACTOR=2.25;}		// Daily time steps
	if (pas>=240) {PERCFACTOR=1;}		// Montly time steps

	//sciprint("PAS=%5.2f PERCFACTOR=%5.2f\n",PAS[0],PERCFACTOR);	


	/* --- Initialisations ----------------- */
	nLAC=0;

	// Parameter vector
	X[0]=Xinp[0];
	X[1]=Xinp[1];
	X[2]=60;		// R reservoir size is fixed
	if(X[0]<1){X[0]=1;}
	if(X[1]<0){X[1]=0;}
	if(X[1]>50){X[1]=50;}
	//printf("X1=%7.2f X2=%7.2f\n",X[0],X[1]);

	if(INI[0]<0){INI[0]=0;}
	if(INI[0]>1){INI[0]=1;}
	if(INI[1]<0){INI[1]=0;}
	if(INI[1]>1){INI[1]=1;}
		
	// State variable vector
	Sini=X[0]*INI[0]; // initial values of S and R
	Rini=X[2]*INI[1];	
	V[0]=Sini;
	V[1]=Rini;
	
	/* ------- Loop over the data -------- */
	for (i=0;i<nval;i++){
		P=Pinp[i];
		E=Einp[i];
		Q=0;

		/*--- Missing data management ---------------------------------*/ 
		if((P<0)&(nLAC<=NLACMAX)){P=0;nLAC++;}
		if((E<0)&(nLAC<=NLACMAX)){E=0;nLAC++;}
		if(nLAC>NLACMAX){V[0]=Sini;V[1]=Rini;P=0;E=0;nLAC=0;}
	
		/*--- CORE MODEL CODE -----------------------------------------*/ 
		
		// ......Production.............
    phi = tanh(P/X[0]);
    S1 = (V[0]+X[0]*phi)/(1+phi*V[0]/X[0]);
    P1 = P + V[0]-S1;
    //printf("[%4d] %7.2f|%7.2f",i,P1,S1);
    
		// ......Percolation.............
    psi = tanh(E/X[0]);
		S2=S1*(1-psi)/(1+psi*(1-S1/X[0]));
		V[0] = S2/pow(1+pow(S2/X[0],3),0.33333);
		P2=S2-V[0];		
    //printf(" %7.2f|%7.2f (%7.2f)\n",P2,S2,pow(1+pow(S2/X[0],3),0.33333));

		// ......Routing store and Water exchange ..
		P3=P1+P2;				
		R1 = V[1]+P3;
		ECH=(X[1]-1)*R1;
		R2=0;IGF=-R1;TPr = R1 + ECH;
		// Case where the reservoir content is not sufficient
		if(TPr>0){R2=TPr;IGF=ECH;}
		Q2 = pow(R2,2)/(R2+X[2]);
		V[1]= R2-Q2;
		
	
		/*--- RESULTS ---------------- */

		Qcalc[i]=Q2;
		Scalc[i]=V[0];
		Rcalc[i]=V[1];
		ECHcalc[i]=ECH;
		PRcalc[i]=P3;
		
		/*--- Missing data management (initial value of S and R)----------*/ 
		if(nLAC<=NLACMAX){Sini=V[0];Rini=V[1];}
	}

}
