#include <math.h>

// Size of X and V vector (WARNING : max UH ordinates = (int)(XSIZE/3)-5 !!)
#define XSIZE 1500

// Number of tolerated missing data
#define NLACMAX 5

// Number of parameters
#define	npX 5

/***********************************************
* GR4J model
*	PAS = Time step (hours)	
*
*	GR4J Parameters :
*	X[0] = Soil reservoir capacity 		(> 1mm)
*	X[1] = losses parameter 		(in [-50..50])
*	X[2] = Routing reseroir capacity	(> 1mm)
*	X[3] = Time constant			(in [0..XSIZE/3-5])
*	X[4] = .. Not utilised ..
*	Ref: 
*	Perrin, C., Michel, C., Andreassian, V.
*	Improvement of a parsimonious model for streamflow simulation
*	(2003) Journal of Hydrology, 279 (1-4), pp. 275-289. Cited 27 times.
************************************************/

/* ************** UH subroutines ***************************** 
*	NPX = Nb of parameters
*	NH  = max length of the UH
*	C   = Time constant of UH
*/
double SS1(double I,double C,double pas)
{
	double Expos;
	
	// Exposant de l'UH
	Expos=2.5;
	if(pas<6){Expos=1.25;}

  if (I<0) {return 0;}
	if (I<C) {return pow(I/C,Expos);}
	if (I>=C){return 1;}
	return 0;
}

double SS2(double I,double C,double pas)
{
	double Expos;
	
	// Exposant de l'UH
	Expos=2.5;
	if(pas<6){Expos=1.25;}
	
	if (I<0)   {return 0;}
	if (I<=C)  {return 0.5*pow(I/C,Expos);}
	if (I<2*C) {return 1-0.5*pow(2-I/C,Expos);}
	if (I>=2*C){return 1;}
	return 0;
}

void UH1 (int NPX, int NH, double * X,double C,double pas)
{
	int i;
	for(i=0;i<NH;i++){X[NPX+i]=SS1((double)(i+1),C,pas)-SS1((double)(i),C,pas);}				
}

void UH2 (int NPX, int NH,double * X,double C,double pas)
{
	int i;
	for(i=0;i<2*NH;i++){X[NPX+NH+i]=SS2((double)(i+1),C,pas)-SS2((double)(i),C,pas);}				
}


/* ************** Core subroutine ****************************** */

void c_GR4J(int nval,double * PAS,double * Xinp,double * Pinp,double * Einp,
  double * Rcalc,double * Scalc,double * Qcalc, double * INI, double * ECHcalc,
  double * ERcalc,double * PRcalc)
{
	int i,k,l,nH,nLAC;
	double P,E,Q,Sini,Rini; 
	double V[XSIZE],X[XSIZE];
	double EN,ES,PS,PR,WS,S2,PERC,ECH,TP,R2,QR,QD,PERCFACTOR,pas;
	double ech1,ech2;

	/* --- Model structure ----------------- */
	//Time step
	pas=PAS[0];
	if((pas<24)&(floor(24/pas)!=24/pas)){for(i=0;i<nval;i++){Qcalc[i]=-9.999;return;}} 
	//printf("pas=%7.0f\n",pas);

	// Maximum UH size	
	if(pas<=6){nH=(int)(XSIZE/3)-npX-1;} 
	if(pas>6){nH=(int)(XSIZE/25);}
	if(pas>=24){nH=(int)(XSIZE/50);}

	// Percolation factor
	if (pas>=12) {PERCFACTOR=2.25;}	// Daily time steps
	if (pas<12) 	{PERCFACTOR=4;} 	// Hourly time steps
	//printf("PAS=%5.2f PERCFACTOR=%5.2f\n",PAS[0],PERCFACTOR);	

	/* --- Initialisations ----------------- */
	nLAC=0;

	// Parameter vector
	for(i=0;i<npX;i++){X[i]=Xinp[i];}

	if(X[0]<1){X[0]=1;}
	if(X[1]<-100){X[1]=-100;}if(X[1]>100){X[1]=100;}
	if(X[2]<1){X[2]=1;}
	if(X[3]<0.5){X[3]=0.5;}if(X[3]>nH){X[3]=nH;}
	for (i=npX;i<XSIZE;i++){X[i]=0;}

	if(INI[0]<=0){INI[0]=0.001;}if(INI[0]>1){INI[0]=1;}
	if(INI[1]<=0){INI[1]=0.001;}if(INI[1]>1){INI[1]=1;}
		
	// State variable vector
	Sini=X[0]*INI[0]; // initial values of S and R
	Rini=X[2]*INI[1];	
	V[0]=Sini;
	V[1]=Rini;
	for (i=2;i<XSIZE;i++){V[i]=0;}
	
	/*--- UH ------------------------------ */
	UH1(npX,nH,X,X[3],pas);
	UH2(npX,nH,X,X[3],pas);

	// Parameter check
	/*
	printf("\nINI:\n");
	printf("INI[0]=%6.4f\n",INI[0]);
	printf("INI[1]=%6.4f\n",INI[1]);
	printf("\nParameters:\n");
	for(i=0;i<4;i++){printf("X[%d]=%6.4f\n",i,X[i]);};
	printf("\nUH1:\n");
	for(i=npX;i<npX+10;i++){printf("X[%d]=%6.4f\n",i,X[i]);};
	printf("\nUH2:\n");
	for(i=npX+nH;i<npX+nH+10;i++){printf("X[%d]=%6.4f\n",i,X[i]);};
	*/

	/* ------- Loop over the data -------- */
	for (i=0;i<nval;i++){
		P=Pinp[i];E=Einp[i];Q=0;

		/*--- Missing data management ---------------------------------*/ 
		if((P<0)&(nLAC<=NLACMAX)){P=0;nLAC++;}
		if((E<0)&(nLAC<=NLACMAX)){E=0;nLAC++;}
		if(nLAC>NLACMAX)         {V[0]=Sini;V[1]=Rini;P=0;E=0;nLAC=0;}
	
		/*--- CORE MODEL CODE -----------------------------------------*/ 
		// ......Production.............
		if(P>E)	{ES=0;
			WS=(P-E)/X[0];
			if(WS>=13){WS=13;}
			PS=X[0]*(1-pow(V[0]/X[0],2))*tanh(WS)/(1+V[0]/X[0]*tanh(WS));
			PR=P-E-PS;
			}
		else	{WS=(E-P)/X[0];
			if(WS>=13){WS=13;}
			ES=V[0]*(2-V[0]/X[0])*tanh(WS)/(1+(1-V[0]/X[0])*tanh(WS));
			PS=0;
			PR=0;
			EN=E-P;
		}
		V[0]+=PS-ES;

		// ......Percolation.............
		S2=V[0]/pow(1+pow(V[0]/PERCFACTOR/X[0],4),0.25);
		PERC=V[0]-S2;
		V[0]=S2;
		PR+=PERC;

		// ......UH1 ....................		
		for (k=0;k<nH-1;k++){
			V[2+k]=V[3+k]+X[npX+k]*PR;
		}
		V[2+nH-1]=X[npX+nH-1]*PR;
				
		// ......UH2 ....................		
		for (l=0;l<2*nH-1;l++){
			V[2+nH+l]=V[3+nH+l]+X[npX+nH+l]*PR;
		}
		V[2+3*nH-1]=X[npX+3*nH-1]*PR;

		// ......Potential Water exchange ....................		
		ECH=X[1]*pow(V[1]/X[2],3.5);

		// ......Routing store calculation..........		
		TP=V[1]+V[2]*0.9+ECH;

		// Case where Reservoir content is not sufficient
		ech1=-V[1]-V[2]*0.9;V[1]=0; 

		if(TP>=0){V[1]=TP;ech1=ECH;}
		//printf("[%5d] %8.2f %8.2f %8.2f  ",i,TP,ECH,ech1);

		R2=V[1]/pow(1+pow(V[1]/X[2],4),0.25);
		QR=V[1]-R2;
		V[1]=R2;
		
		// ......Direct runoff calculation.........	
		QD=0;

		// Case where the UH cannot provide enough water
		TP=V[2+nH]*0.1+ECH;
		ech2 = -V[2+nH]*0.1;QD=0;
		if(TP>0){QD=V[2+nH]*0.1+ECH;ech2=ECH;}
		//printf("%8.2f %8.2f %8.2f\n",V[2+nH]*0.1+ECH,ECH,ech2);

		// ......TOTAL STREAMFLOW..................
		Q=QD+QR;
	
		/*--- RESULTS ---------------- */
		Qcalc[i]=Q;
		Scalc[i]=V[0];
		Rcalc[i]=V[1];
		ECHcalc[i]=ech1+ech2;
		ERcalc[i]=ES+EN;
		PRcalc[i]=PR;
		
		/*--- Missing data management (initial value of S and R)----------*/ 
		if(nLAC<=NLACMAX){Sini=V[0];Rini=V[1];}
	}
}
