/*
* Daily model : 
* Calculation of nearest neighbours according the methodology detailed  
* by Lall, U., Sharma, A., 1996. A nearest neighbour bootstrap for time 
* series resampling. Water Resources Research 32 (3), 679693.
* 
* Inputs 
*   NBK   Number of nearest neighbours to consider
*   WINS  Temporal window to restrain the search for nearest neighbours
*   DMAT  Matrix of feature vectors for each days (nvalDMAT x ncolDMAT)
*   WEI   Weights to calculate the euclidian distance (ncolDMAT x 1)
*   SMAT  Data to resample
*   NNdep Initial day
*   RND   Random number vector
*   
* Outputs
*   KNNSIM Resampled data
*/
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <sciprint.h>

// Maximum number of matrix size
#define SIZEMAX 200000

// Maximum number of nearest neighbour
#define MAXCOL 100

// Maximum value of the distance
#define MAXVALDIST 1000

// Function to sort
void SORTORD(int nvalDMAT, double * IN, int * indexORD);

void c_KNN1SIM(int nvalDMAT, int ncolDMAT,int nrand, double * NBK, 
  double * WINS, double * DMAT, double * WEI, double * DMATSELEC, 
  double * NNdep, double * RND, double * KNNSIM_NN)
{
    int nbk,dmatCol,nnCol,day,day2,NbAnn,ann,debDay,finDay,nnSelec,nDep;
    int NNliste[MAXCOL],nnClass,nbjDMATSELEC;
    double DIST[MAXCOL],CARAC[MAXCOL],winlag,SomW;
    double rnd, KERN[MAXCOL],sumkern,DISTtemp;

    /*----- Parameters check ----------*/
    // Error if the number of days is greater than SIZEMAX
    // or if the number of features is greater than NBFEATMAX
    if(nvalDMAT*ncolDMAT>=SIZEMAX){exit(1);}
    
    nbk=(int)(NBK[0]);
    if(nbk>=MAXCOL){exit(1);}
    
    // Half temporal window
    winlag=WINS[0]/2;if(winlag<=0){exit(1);}

    // Number of years in DMAT and SMAT 
    NbAnn = (int)(nvalDMAT/365.25)+1;

    sciprint("\n\tKNN parameters: NBK=%2d WIN=%2.0f DMAT[%5d %2d],NVALSIM=%5d ",nbk,WINS[0],nvalDMAT,ncolDMAT,nrand);

    // Normalisation of weights
    SomW=0;
    for(dmatCol=0;dmatCol<ncolDMAT;dmatCol++){SomW+=fabs(WEI[dmatCol]);}
    sciprint("\n\tVariables weights:\n\t");
    if(SomW>0){
      for(dmatCol=0;dmatCol<ncolDMAT;dmatCol++){
        WEI[dmatCol]=fabs(WEI[dmatCol])/SomW;
        sciprint("%3.1f ",WEI[dmatCol]);
        }
    }

    // Nb de jours slectionns
    nbjDMATSELEC=1;
    for(day=0;day<nvalDMAT;day++){if(DMATSELEC[day]>0){nbjDMATSELEC++;}}
    sciprint("\n\tPercentage of selected days: %6d over %6d (%2.0f perc)",nbjDMATSELEC,nvalDMAT,(double)(nbjDMATSELEC)/(double)(nvalDMAT)*100);
    if(DMATSELEC[(int)(NNdep[0])]<=0){sciprint("CAUTION ! The starting day is not a selected day >> problem with the resampling algo.");}
    
    /*----- Initialisation des caractristiques ----------*/
    nDep=(int)(NNdep[0]);if((nDep<0)|(nDep>=nvalDMAT)){nDep=0;}
    for(dmatCol=0;dmatCol<ncolDMAT;dmatCol++){CARAC[dmatCol]=DMAT[nDep+nvalDMAT*dmatCol];}
    
    /*----- Resampling kernel ----------*/
    sumkern=0;for(nnCol=0;nnCol<nbk;nnCol++){sumkern+=1/(double)(nnCol+1);}
    KERN[0]=1/sumkern;
    sciprint("\n\tKernel = %5.3f ",KERN[0]);
    for(nnCol=1;nnCol<nbk;nnCol++){
      KERN[nnCol]=KERN[nnCol-1]+1/(double)(nnCol+1)/sumkern;
      sciprint("%5.3f ",KERN[nnCol]);
    }
    sciprint("\n\tLine nb:\n\t");
        
    /*----- Resampled data generation ----------*/
    for(day=0;day<nrand;day++){    
        if(fmod(day,1000)==0){sciprint("%5d ",day);}
        if(fmod(day,8000)==0){sciprint("\n\t");}
                
        // initialize the distance vector
        for(nnCol=0;nnCol<nbk;nnCol++){DIST[nnCol]=MAXVALDIST;}

        // Calculate the distance with the value of the descriptors
        // calculate only if the current date falls within the time window
        for(ann=-1;ann<NbAnn;ann++){
            // Start and end of the temporal window for each year
            debDay=(int)(ann*365.25-winlag+fmod(day,365.25));
            finDay=(int)(ann*365.25+winlag+fmod(day,365.25));
            if(debDay<0){debDay=0;}
            if(finDay<0){finDay=0;}   
            if(debDay>=nvalDMAT){debDay=nvalDMAT-1;} 
            if(finDay>=nvalDMAT){finDay=nvalDMAT-1;}    
            //printf("A%2d(%5d<>%5d) ",ann,debDay,finDay);
                    
            for(day2=debDay;day2<finDay;day2++){
                // Parcours la base de donnes en ne retenant que les jours slectionns
                if(DMATSELEC[day2]>0){
                    DISTtemp=0;
                    // Weighted euclidian distance
                    for(dmatCol=0;dmatCol<ncolDMAT;dmatCol++){
                        DISTtemp+=WEI[dmatCol]*pow(DMAT[day2+nvalDMAT*dmatCol]-CARAC[dmatCol],2);
                    }
                    // Check if the distance is lower than one of the already stored distance
                    nnClass=0;
                    while((DISTtemp>DIST[nnClass])&(nnClass<=nbk)){nnClass++;}
                    // if yes, rearrange DIST and NNliste vectors 
                    if(nnClass<nbk){
                        for(nnCol=nbk-1;nnCol>=nnClass+1;nnCol--){
                            DIST[nnCol]=DIST[nnCol-1];NNliste[nnCol]=NNliste[nnCol-1];
                        }
                        NNliste[nnClass]=day2;DIST[nnClass]=DISTtemp;
                    }
                } 
            }
        }
      
        //printf("** [%5d] ",day);
        //for(nnCol=0;nnCol<nbk;nnCol++){printf(" %5d",NNliste[nnCol]+1);}

        // Random number
        rnd=RND[day];if((rnd<0)|(rnd>1)){rnd=(double)(rand())/RAND_MAX;}

        // Selection of the nearest neighbour
        nnCol=0;while((rnd>KERN[nnCol])&(nnCol<nbk)){nnCol++;}
        
        nnSelec = (int)(NNliste[nnCol]+1);
        // Cas o le plus proche voisin est le dernier lment
        if(nnSelec>=nvalDMAT){nnSelec=nvalDMAT-1;}
        
        // Stockage des caractristiques du jour
        for(dmatCol=0;dmatCol<ncolDMAT;dmatCol++){CARAC[dmatCol]=DMAT[nnSelec+nvalDMAT*dmatCol];}

        // Attribution of the data to KNNSIM matrix
        KNNSIM_NN[day]=(double)(nnSelec)+1;
        //printf(" >> %5.0f\n",KNNSIM_NN[day]);
        
    } // End of the simulation
    
    sciprint("\n\n");
    
}
